from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship

from pyrspb.objects.database import Base


class Material(Base):
    __tablename__ = "materials"
    name = Column(String, nullable=False, primary_key=True)
    level = Column(Integer, nullable=False)
    xp = Column(Integer, nullable=False)
    rarity = Column(String, nullable=False)
    short_sources = Column(String, nullable=False)
    full_list_url = Column(String, nullable=False)
    sources = relationship("MaterialSource", backref="materials")

    def __str__(self):
        return "{}\n" \
               "Level: {}\n" \
               "Xp: {}\n" \
               "Sources: {}\n" \
               "Full list: {}".format(self.name, self.level, self.xp, self.short_sources, self.full_list_url)

    def __repr__(self):
        return f"<Material(name={self.name}, level={self.level}, xp={self.xp}," \
               f" rarity={self.rarity}, short_sources={self.short_sources}, full_list_url={self.full_list_url})>"



from contextlib import closing
from bs4 import BeautifulSoup
from requests import get
from requests.exceptions import RequestException


def get_page(url):
    try:
        with closing(get(url, stream=True)) as resp:
            if resp.status_code == 200:
                return BeautifulSoup(resp.content, "html.parser")
            else:
                return None
    except RequestException as e:
        print("Error @ \"{}\": {}".format(url, str(e)))
        return None


class PageParser:
    def __init__(self, location):
        self.url = location
        self.content = get_page(self.url)

    def __parse(self):
        pass



from typing import List

from bs4.element import PageElement

from pyrspb.objects import PageParser, Material
from pyrspb.objects.database import session
from pyrspb.utils.string_formatting import *
from . import MaterialSourceScraper


def select_relevant_columns(row: List[PageElement]):
    _, _, *columns = row
    return columns


def get_rows(table):
    _, *rows = table.find_all("tr")
    return [select_relevant_columns(row) for row in rows]


def extract_info(rarity, row):
    count = len(row)
    name_element, level_element, xp_element, *rest = row
    data = []

    if count == 5:
        sources_element, full_list_element = rest
        data.append(sources_element.text.strip())
        data.append("https://runescape.wikia.com" + full_list_element.a["href"])
    elif count == 4:
        sources = "See the full list of items for source info"
        full_list_element = rest[-1]
        data.append(sources)
        data.append("https://runescape.wikia.com" + full_list_element.a["href"])
    else:
        return None

    sources, full_list = data
    name = name_element.text.strip()
    level = int(level_element.text)
    xp = make_float(xp_element.text)

    return Material(name=name,
                    level=level,
                    xp=xp,
                    short_sources=sources,
                    full_list_url=full_list,
                    rarity=rarity)


class MaterialScraper(PageParser):
    def __init__(self):
        PageParser.__init__(self, "https://runescape.wikia.com/wiki/Materials")
        self.materials = self.__parse()

    def __parse(self):
        common_table, uncommon_table, rare_table, *_ = self.content.find_all("table")
        self.common = [extract_info("Common", row) for row in get_rows(common_table)]
        self.uncommon = [extract_info("Uncommon", row) for row in get_rows(uncommon_table)]
        self.rare = [extract_info("Rare", row) for row in get_rows(rare_table)]
        return self.common + self.rare + self.uncommon

    def save_to_db(self):
        session.query(Material).delete()
        # for m in self.materials:
        #
        #     session.query(Material).filter(Material.name == m.name).delete()
        #     session.add(new_material)
        session.add_all(self.materials)
        session.commit()



import re

from pyrspb.objects.page_parser import PageParser


class VersionScraper(PageParser):
    def __init__(self):
        PageParser.__init__(self, "https://rspocketbook.com/")
        self.version, self.date = self.__parse()

    def __parse(self):
        *junk, version = self.content.find_all("div", class_="box")
        full_version_string = version.find_all("p")[-1].text
        # v
        version, date = re.search('(v[\d|\.]+\d). (.+)', full_version_string).groups()
        return [version, date]




import datetime

from sqlalchemy import Column, Integer, String, Float, ForeignKey, DateTime
from sqlalchemy.orm import relationship

from pyrspb.objects.database import Base
from pyrspb.objects.invention.material import Material


class MaterialSource(Base):
    __tablename__ = "material_sources"

    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    materials_each = Column(Integer, nullable=False)
    junk_chance = Column(Float, nullable=False)
    price_per_disassemble = Column(Float, nullable=False)
    buy_limit = Column(Integer, nullable=False)
    raw_chance = Column(Float, nullable=False)
    average_number = Column(Float, nullable=False)
    cost_per_material = Column(Float, nullable=False)
    materials_per_hour = Column(Float, nullable=False)
    material_name = Column(String, ForeignKey("materials.name"))
    material = relationship(Material, backref="material_sources")
    last_updated = Column(DateTime, onupdate=datetime.datetime.now, default=datetime.datetime.now())

    def __str__(self):
        return "{} ({}/4hrs) gives {} {} per hour at a cost of {} gp each".format(self.name,
                                                                                  self.buy_limit,
                                                                                  self.materials_per_hour,
                                                                                  self.material.name,
                                                                                  self.cost_per_material)

    def __repr__(self):
        return f"<MaterialSource(name={self.name})>"

#
#
# class InventionItem:
#     def __init__(self, columns):
#         item_name, materials_each, junk_chance, \
#           price_per_disassemble, buy_limit, raw_chance, \
#           average_number, cost_per_material, materials_per_hour = columns
#
#         self.item_name = item_name.a["title"]
#         self.materials_each = make_int(materials_each.string)
#         self.junk_chance = percent_str_to_float(junk_chance.string)
#         self.price_per_disassemble = make_float(price_per_disassemble.string)
#         self.buy_limit = make_int(buy_limit.string)
#         self.raw_chance = percent_str_to_float(raw_chance.string)
#         self.average_number = percent_str_to_float(average_number.string)
#         self.cost_per_material = make_float(cost_per_material.string)
#         self.materials_per_hour = make_float(materials_per_hour.string)
#
#     def __str__(self):
#         return "{} ({}/4hrs) gives {} materials per hour at a cost of {} gp each".format(self.item_name,
#                                                                                          self.buy_limit,
#                                                                                          self.materials_per_hour,
#                                                                                          self.cost_per_material)



from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
import os
path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "pyrspb.db")
engine = create_engine(f"sqlite:////{path}", echo=False)
Base = declarative_base()
Base.metadata.bind = engine
DBSession = sessionmaker(bind=engine)
session = DBSession()



import datetime
from typing import List

from pyrspb.objects.database import session
from pyrspb.objects.invention import Material, MaterialSource
from pyrspb.parsers.invention import MaterialSourceScraper


def material_from_string(s: str) -> Material:
    return session.query(Material).filter(Material.name.ilike(f"%{s}%")).one_or_none()


def identify_ideal_sources(m: Material, buy_limit=100, cost_per=100000, per_hour=100) -> List[MaterialSource]:
    # Save sources if they don't exist
    if m.sources is None:
        MaterialSourceScraper(m).save_to_db()

    sources: List[MaterialSource] = m.sources
    # Last time data was pulled
    last_updated: datetime.timedelta = datetime.datetime.now() - sources[0].last_updated
    # Update the db if it's been over a day
    if last_updated.days >= 1:
        MaterialSourceScraper(m).save_to_db()

    return session.query(MaterialSource).filter(MaterialSource.buy_limit >= buy_limit) \
        .filter(MaterialSource.cost_per_material <= cost_per) \
        .filter(MaterialSource.materials_per_hour >= per_hour) \
        .order_by(MaterialSource.cost_per_material).all()



from pyrspb.parsers.invention import MaterialScraper, MaterialSourceScraper
import json
from pyrspb.objects.invention.material import Material
from pyrspb.objects.database import Base, engine, session
from pyrspb.api.invention import material_calculator
import os

print(Material.__table__)
Base.metadata.create_all(engine)
MaterialScraper().save_to_db()
x = MaterialSourceScraper(session.query(Material).filter(Material.name.ilike("%direct%")).one())
items = x.sources
for item in items:
    print(item)
    print(item.material)
x.save_to_db()

# calc = MaterialCalculator("Direct")
# for item in reversed(calc.ideal_items):
#     print(item)
# for s in material_calculator.identify_ideal_sources(material_calculator.material_from_string("direct")):
#     print(s)
# mats = MaterialsScraper()
# mats.dump_to_json()
# for mat in mats.materials:
#     print(str(Material(mat.rarity, dict=mat.to_dict())))
# print("https://discordapp.com/oauth2/authorize?client_id={}&scope=bot".format("470682754929917963"))

# book = BookVersion()
# print(book.version, "|", book.date)

# for item in MaterialCalculator("Dextrous").ideal_items[:5]:
#     print(str(item))

# MaterialCalculator("Dextrous")
# for i in MaterialCalculator("Dextrous").ideal_items[:5]:
#     print(i)
#     print("------")
# mats = MaterialCalculator("Direct")
#
# for i in mats.ideal_items[:5]:
#     print(json.dumps(i.__dict__))
#     print(i)
# for i in mats.ideal_items:
# print("YO", i)


# print(session.query(Material).all())
# direct: Material = session.query(Material).filter(Material.name.ilike("%direct%")).one()
# for source in direct.sources:
#     print(source)



import re


def percent_str_to_float(s):
    try:
        res = re.match("""(\d+\.?\d*)""", s)
        if res is not None:
            return round(float(res.group(0)) / 100.0, 2)
        else:
            return -1
    except TypeError:
        return -1


def make_float(s):
    try:
        return float(s.replace(',', ''))
    except ValueError:
        return -1


def make_int(s):
    try:
        return int(s.replace(',', ''))
    except ValueError:
        return -1



from pyrspb.objects import MaterialSource, PageParser
from pyrspb.objects.database import session
from pyrspb.utils.string_formatting import *


def columns_to_item(material, cols):
    item_name_element, materials_each_element, junk_chance_element, \
    price_per_disassemble_element, buy_limit_element, raw_chance_element, \
    average_number_element, cost_per_material_element, materials_per_hour_element = cols

    item_name = item_name_element.a["title"]
    materials_each = make_int(materials_each_element.string)
    junk_chance = percent_str_to_float(junk_chance_element.string)
    price_per_disassemble = make_float(price_per_disassemble_element.string)
    buy_limit = make_int(buy_limit_element.string)
    raw_chance = percent_str_to_float(raw_chance_element.string)
    average_number = percent_str_to_float(average_number_element.string)
    cost_per_material = make_float(cost_per_material_element.string)
    materials_per_hour = make_float(materials_per_hour_element.string)

    return MaterialSource(name=item_name,
                          materials_each=materials_each,
                          junk_chance=junk_chance,
                          price_per_disassemble=price_per_disassemble,
                          buy_limit=buy_limit,
                          raw_chance=raw_chance,
                          average_number=average_number,
                          cost_per_material=cost_per_material,
                          materials_per_hour=materials_per_hour,
                          material=material)


class MaterialSourceScraper(PageParser):
    def __init__(self, material):
        self.material = material
        PageParser.__init__(self, self.__get_material_url())
        self.sources = self.__parse()

    def __get_material_url(self):
        return "https://runescape.wikia.com/wiki/Calculator:Disassembly_by_material/{}" \
            .format(self.material.name.split(" ")[0])

    def __parse(self):
        table = self.content.find_all("tr")
        columns = [list(row.children) for row in table]
        pred = lambda c: len(c) > 3 and 'th' not in list(map(lambda n: n.name, c))
        rows = list(filter(pred, columns))

        return [columns_to_item(self.material, row) for row in rows]

    def save_to_db(self):
        query = session.query(MaterialSource).filter(MaterialSource.material == self.material)
        map(session.delete, query)
        session.add_all(self.sources)
        session.commit()



from setuptools import find_packages, setup

setup(
    version='0.0.2.2',
    name="pyrspb",
    packages=find_packages(),
    author="RS Pocketbook",
    author_email="admin@rspocketbook.com",
    url="https://gitlab.com/RS_PocketBot/pyrspb",
    install_requires=[
        "beautifulsoup4",
        "requests",
        "sqlalchemy"
    ],
    include_package_data=True
)
